/*
 * programming the dpl3518a Dolby Pro Logic Processor
 *
 * WARNING: THIS DRIVER WILL LOAD WITHOUT COMPLAINTS EVEN IF A DIFFERENT
 * CHIP IS AT ADDRESS 0x84 (it relies on i2c to make sure that there is a
 * device acknowledging that address)
 *
 *  Copyright (C) 1999 Roland Jansen <roland@lut.rwth-aachen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * ---
 * DOLBY and PRO LOGIC are trademarks of 
 * Dolby Laboratories Licensing Corporation.
 */

/* FIXME */
#define DPL_MAJOR  127

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <asm/semaphore.h>
#ifdef __SMP__
#include <asm/pgtable.h>
#include <linux/smp_lock.h>
#endif

#include <linux/unistd.h>

#include "aiee.h"

#include "i2c.h"
#include "dpl3518.h"

/* insmod parameters */
static int debug = 0;		/* debug output */

struct dpl3518 {
	struct i2c_bus *bus;
	int mode;		/* dolby mode */

};

//#define MAX_DSP_DEVICES 1
//static struct dpl3518 *dpl_device[MAX_DSP_DEVICES];
static struct dpl3518 *dpl_device;

/* ---------------------------------------------------------------------- */

#define dprintk     if (debug) printk

#if LINUX_VERSION_CODE > 0x020100
MODULE_PARM(debug, "i");
#endif

/* ---------------------------------------------------------------------- */

#define I2C_DPL3518       0x84
#define I2C_DPL3518_MR    0x10	/* write address MODE_REG */
#define I2C_DPL3518_DFP   0x12	/* write address DFP */

/* ----------------------------------------------------------------------- */
/* functions for talking to the dpl3518 Sound processor                   */

static int dpl3518_reset(struct i2c_bus *bus)
{
	int ret = 0;

	mdelay(2);
	i2c_start(bus);
	i2c_sendbyte(bus, I2C_DPL3518, 2000);
	i2c_sendbyte(bus, 0x00, 0);
	i2c_sendbyte(bus, 0x80, 0);
	i2c_sendbyte(bus, 0x00, 0);
	i2c_stop(bus);
	mdelay(2);
	i2c_start(bus);
	if (0 != i2c_sendbyte(bus, I2C_DPL3518, 2000) ||
	    0 != i2c_sendbyte(bus, 0x00, 0) ||
	    0 != i2c_sendbyte(bus, 0x00, 0) ||
	    0 != i2c_sendbyte(bus, 0x00, 0)) {
		ret = -1;
		printk(KERN_ERR "dpl3518: chip reset failed, penguin on i2c bus?\n");
	}
	i2c_stop(bus);
	mdelay(2);
	return ret;
}

static int dpl3518_read(struct i2c_bus *bus, int dev, int addr)
{
	int err, ret;
	short val = 0;

	for (err = 0; err < 3;) {
		ret = 0;
		i2c_start(bus);
		if (0 != i2c_sendbyte(bus, I2C_DPL3518, 2000) ||
		    0 != i2c_sendbyte(bus, dev + 1, 0) ||
		    0 != i2c_sendbyte(bus, addr >> 8, 0) ||
		    0 != i2c_sendbyte(bus, addr & 0xff, 0)) {
			ret = -1;
		} else {
			i2c_start(bus);
			if (0 != i2c_sendbyte(bus, I2C_DPL3518 + 1, 2000)) {
				ret = -1;
			} else {
				val |= (int) i2c_readbyte(bus, 0) << 8;
				val |= (int) i2c_readbyte(bus, 1);
			}
		}
		i2c_stop(bus);
		if (0 == ret)
			break;

		/* some I/O error */
		err++;
		printk(KERN_WARNING "dpl3518: I/O error #%d (read 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ / 10);
	}
	if (-1 == ret) {
		printk(KERN_WARNING "dpl3518: giving up, reseting chip. DPL will go off, sorry folks :-|\n");
		dpl3518_reset(bus);
	}
	return val;
}

static int dpl3518_write(struct i2c_bus *bus, int dev, int addr, int val)
{
	int ret, err;

	for (err = 0; err < 3;) {
		ret = 0;
		i2c_start(bus);
		if (0 != i2c_sendbyte(bus, I2C_DPL3518, 2000) ||
		    0 != i2c_sendbyte(bus, dev, 0) ||
		    0 != i2c_sendbyte(bus, addr >> 8, 0) ||
		    0 != i2c_sendbyte(bus, addr & 0xff, 0) ||
		    0 != i2c_sendbyte(bus, val >> 8, 0) ||
		    0 != i2c_sendbyte(bus, val & 0xff, 0))
			ret = -1;
		i2c_stop(bus);
		if (0 == ret)
			break;

		/* some I/O error */
		err++;
		printk(KERN_WARNING "dpl3518: I/O error #%d (write 0x%02x/0x%02x)\n",
		       err, dev, addr);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ / 10);
	}
	if (-1 == ret) {
		printk(KERN_WARNING "dpl3518: giving up, reseting chip. DPL will go off, sorry folks :-|\n");
		dpl3518_reset(bus);
	}
	return ret;
}

/* ----------------------------------------------------------------------- */

/* in: 0..100 out: 0x00..0x7f */
static int p100_to_7f(int i)
{
	int r;

	r = ((i & 0x7f) * 127 + 1) / 100;
	if (r > 127)
		r = 127;
	if (r < 0)
		r = 0;
	return r;
}

/* ----------------------------------------------------------------------- */

void dpl3518_ioc_init(struct i2c_bus *bus)
{
	dprintk("dpl3518 init\n");
	dpl3518_write(bus, I2C_DPL3518_MR, 0x0083, 0x0008); /* mode_reg */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x000b, 0x0320); /*I2S1 src: RL */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0012, 0x0000); /*prescale2: off */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0016, 0x1000); /*prescale: 0 dB */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0038, 0x0420); /*I2S2 src: CS */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0040, 0x0000); /*adaptive,surround */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0041, 0x0520); /*i2s1, stereo */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0042, 0x1400); /*delay: 20ms */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0043, 0x0000); /*inp. Bal. auto */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0044, 0x4000); /*spatial 50% */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0045, 0x0000); /*panorama off */
	//dpl3518_write(bus,I2C_DPL3518_DFP,0x0045,0x5400); /*panorama 66%*/
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0046, 0x0000); /*reverb off */
}

void dpl3518_prologic_init(struct i2c_bus *bus)
{
	dprintk("dpl3518 prologic_init\n");
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x000b, 0x0320); /*I2S1 src: RL */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0038, 0x0420); /*I2S2 src: CS */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0041, 0x0520); /*i2s1, stereo */
}

void dpl3518_noise_init(struct i2c_bus *bus)
{
	dprintk("dpl3518 noise_init\n");
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x000b, 0x0320); /*I2S1 src: RL */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0038, 0x0420); /*I2S2 src: CS */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0041, 0x0100); /*NOISE */
}

void dpl3518_through_init(struct i2c_bus *bus)
{
	dprintk("dpl3518 through_init\n");
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x000b, 0x0520); /*I2S1 src: S1 */
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0038, 0x0220); /*I2S2 src: scart*/
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0041, 0x0520); /*i2s1, stereo */
}

void dpl3518_setmode(struct i2c_bus *bus, int mode)
{
	dprintk("dpl3518 setmode: %i\n", mode);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0040, mode << 4);
	dpl_device->mode = mode;
}

void dpl3518_setnoisemode(struct i2c_bus *bus, int mode)
{
	static int noise_modes[5] = { 0x0000, 0x01a0, 0x01b0, 0x01c0, 0x01d0 };
	dprintk("dpl3518 setnoisemode: %i\n", mode);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0041, noise_modes[mode]);
}


void dpl3518_delay(struct i2c_bus *bus, int delay)
{
	dprintk("dpl3518 setdelay: %i\n", delay);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0042, delay << 8);
}

void dpl3518_reverb(struct i2c_bus *bus, int rev)
{
	dprintk("dpl3518 setreverb: %i\n", rev);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0046, p100_to_7f(rev) << 8);
}

void dpl3518_panorama(struct i2c_bus *bus, int pan)
{
	dprintk("dpl3518 panorama: %i\n", pan);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0045, p100_to_7f(pan) << 8);
}

void dpl3518_spatial(struct i2c_bus *bus, int spatial)
{
	dprintk("dpl3518 setspatial: %i\n", spatial);
	dpl3518_write(bus, I2C_DPL3518_DFP, 0x0045, p100_to_7f(spatial) << 8);
}

/* ----------------------------------------------------------------------- */

static int dpl3518_attach(struct i2c_device *device)
{

	struct dpl3518 *dpl;
	LOCK_FLAGS;

	device->data = dpl = kmalloc(sizeof(struct dpl3518), GFP_KERNEL);
	if (NULL == dpl)
		return -ENOMEM;
	memset(dpl, 0, sizeof(struct dpl3518));
	dpl->bus = device->bus;
	dpl->mode = 0;		/* DOLBY_MODE_NONE */

	LOCK_I2C_BUS(dpl->bus);
	if (-1 == dpl3518_reset(dpl->bus)) {
		UNLOCK_I2C_BUS(dpl->bus);
		kfree(dpl);
		dprintk("dpl3518: no chip found\n");
		return -1;
	}
	UNLOCK_I2C_BUS(dpl->bus);

	strcpy(device->name, "DPL3518");

	printk(KERN_INFO "dpl3518: init\n");

	dpl_device = dpl;
	dpl3518_ioc_init(dpl_device->bus);
	/* done */
	return 0;
}

static int dpl3518_detach(struct i2c_device *device)
{
	struct dpl3518 *dpl = (struct dpl3518 *) device->data;
	LOCK_FLAGS;

	LOCK_I2C_BUS(dpl->bus);
	dpl3518_reset(dpl->bus);
	UNLOCK_I2C_BUS(dpl->bus);
	
	dpl_device = NULL;
	kfree(dpl);
	return 0;
}

static int dpl3518_dev_ioctl(struct inode *inode, struct file *file,
			     unsigned int cmd, unsigned long arg)
{
//        unsigned int minor = MINOR(inode->i_rdev);
	int ret = 0;
	int ver = DPL_VERSION_CODE;

	switch (cmd) {
	case DPL_IOC_RESET:
		dpl3518_reset(dpl_device->bus);
		break;
	case DPL_IOC_VERSION:
#if 0
		if (put_user(DPL_VERSION_CODE, (int *) arg))
			ret = -EFAULT;
#else
		if(copy_to_user((void*)arg,&ver,sizeof(int)))
			return -EFAULT;
#endif
		break;
	case DPL_IOC_INIT:
		dpl3518_ioc_init(dpl_device->bus);
		break;
	case DPL_IOC_PROLOGIC_INIT:
		dpl3518_prologic_init(dpl_device->bus);
		break;
	case DPL_IOC_NOISE_INIT:
		dpl3518_noise_init(dpl_device->bus);
		break;
	case DPL_IOC_THROUGH_INIT:
		dpl3518_through_init(dpl_device->bus);
		break;
	case DPL_IOC_MODE:
		dpl3518_setmode(dpl_device->bus, (int) arg);
		break;
	case DPL_IOC_GET_MODE:
#if 0
		if (put_user(dpl_device->mode, (int *) arg))
			ret = -EFAULT;
#else
		if(copy_to_user((void*)arg,&(dpl_device->mode),sizeof(int)))
			return -EFAULT;
#endif
		break;
	case DPL_IOC_NOISE_MODE:
		dpl3518_setnoisemode(dpl_device->bus, (int) arg);
		break;
	case DPL_IOC_DELAY:
		dpl3518_delay(dpl_device->bus, (int) arg);
		break;
	case DPL_IOC_REVERB:
		dpl3518_reverb(dpl_device->bus, (int) arg);
		break;
	case DPL_IOC_PANORAMA:
		dpl3518_panorama(dpl_device->bus, (int) arg);
		break;
	case DPL_IOC_SPATIAL:
		dpl3518_spatial(dpl_device->bus, (int) arg);
		break;
	default:
		dprintk("dpl3518_ioctl: default %i\n", (int) arg);
		ret = -EINVAL;
	}
	return ret;
}

static int dpl3518_dev_open(struct inode *inode, struct file *file)
{
	dprintk("dpl3518_dev_open\n");
	MOD_INC_USE_COUNT;
	return 0;
}

static int dpl3518_dev_release(struct inode *inode, struct file *file)
{
	dprintk("dpl3518_dev_release\n");
	MOD_DEC_USE_COUNT;
	return 0;
}

/* ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */

struct i2c_driver i2c_driver_dpl =
{
	"dpl3518",			/* name       */
	4,				/* ID         */ /* FIXME */
	I2C_DPL3518, I2C_DPL3518,	/* addr range */

	dpl3518_attach,
	dpl3518_detach,
	NULL
};

static struct file_operations dpl3518_fops =
{
	NULL,			/* llseek               */
	NULL,			/* read                 */
	NULL,			/* write                */
	NULL,			/* readdir              */
	NULL,			/* poll                 */
	dpl3518_dev_ioctl,
	NULL,			/* mmap     */
	dpl3518_dev_open,
	NULL,
	dpl3518_dev_release,
};


#ifdef MODULE

int init_module(void)
#else
int dpl3518_init(void)
#endif
{
	if (register_chrdev(DPL_MAJOR, "dpl3518", &dpl3518_fops)) {
		printk("dpl3518: unable to get major %d\n", DPL_MAJOR);
		return -EBUSY;
	}
	i2c_register_driver(&i2c_driver_dpl);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_unregister_driver(&i2c_driver_dpl);
	unregister_chrdev(DPL_MAJOR, "dpl3518");
}

#endif
